#include <math.h>
#include <stdio.h>
#include <time.h>
#include <stdint.h>
#include <stdlib.h>

#define likely(x)	__builtin_expect((x), 1)
#define unlikely(x) __builtin_expect((x), 0)
#define ITERATIONS 2000000

//#pragma intrinsic(__rdtsc)
// Printing a Line onto the screen
void line()
{
	printf("%%==========================================%%\n");
}
void wline()
{
	printf("%%------------------------------------------%%\n");
}

//uint64_t rdtsc()
//{
//    return __rdtsc();
//}

// Read the instruction-counter of the executing cpu
// No guarantee this will work on amd
uint64_t rdtsc(){
   unsigned int lo,hi;
    __asm__ __volatile__ ("rdtsc" : "=a" (lo), "=d" (hi));
    return ((uint64_t)hi << 32) | lo;
}

// Method without parameters and return
void method()
{

}

// Method with return
int methodRet()
{
	return 1;
}

// Method with param
void methodParam(int param)
{

}

// Method with pointer Parameter
void methodPointer(int *param)
{

}

// Method with big params as array
int methodBig(int a[])
{
	return 42;
}

// Method with params and return
int methodParamRet(int param)
{
	return 42;
}

static inline float ownSqroot(double par)
{
	float i,j = 0;
	float t1, t2;
	while (i*i <= par)
	{
		i += 0.1;
	}

	t1 = i;

	for(j=0;j<10;j++)
	{
		t2 = par;
		t2 /= t1;
		t2 += t1;
		t2 /= 2;
		t1 = t2;
	}

	return t2;
}

// Method for common math operations
void mathFunctions()
{
	uint64_t a, b;
	double c;
	float e = 1.1;

	int i;
	
	c = e; // damit der Compiler nicht meckert

	printf("%% Sqrt\n");
	a = rdtsc();
	for(i = 0; i < ITERATIONS; ++i)
	{	 
		e = sqrt(i);
	}
	b = rdtsc() - a;
	printf("%% CPU Cycles used: %ld\n", b);
	c = (double) b / (double)ITERATIONS;
	printf("%% Cycles per Call %f\n", c);
	wline();

	printf("%% Pow\n");
	a = rdtsc();
	for(i = 0; i < ITERATIONS; ++i)
	{
		e = pow(i,i);
	}
	b = rdtsc() - a;
	printf("%% CPU Cycles used: %ld\n", b);
	c = (double) b / (double)ITERATIONS;
	printf("%% Cycles per Call %f\n", c);
	wline();
	
	printf("%% Exp\n");
	a = rdtsc();
	for(i = 0; i < ITERATIONS; ++i)
	{
			e = exp(i);
	}
	b = rdtsc() - a;
	printf("%% CPU Cycles used: %ld\n", b);
	c = (double) b / (double)ITERATIONS;
	printf("%% Cycles per Call %f\n", c);
	wline();
	
	printf("%% Sin\n");
	a = rdtsc();
	for(i = 0; i < ITERATIONS; ++i)
	{
			e = sin(i);
	}
	b = rdtsc() - a;
	printf("%% CPU Cycles used: %ld\n", b);
	c = (double) b / (double)ITERATIONS;
	printf("%% Cycles per Call %f\n", c);
	wline();
	
	printf("%% Cos\n");
	a = rdtsc();
	for(i = 0; i < ITERATIONS; ++i)
	{
			e = cos(i);
	}
	b = rdtsc() - a;
	printf("%% CPU Cycles used: %ld\n", b);
	c = (double) b / (double)ITERATIONS;
	printf("%% Cycles per Call %f\n", c);
	wline();
	
	printf("%% Tan\n");
	a = rdtsc();
	for(i = 0; i < ITERATIONS; ++i)
	{
			e = tan(i);
	}
	b = rdtsc() - a;
	printf("%% CPU Cycles used: %ld\n", b);
	c = (double) b / (double)ITERATIONS;
	printf("%% Cycles per Call %f\n", c);
	wline();
	
	printf("%% Max\n");
	a = rdtsc();
	for(i = 0; i < ITERATIONS; ++i)
	{
			e = fmax(i, i);
	}
	b = rdtsc() - a;
	printf("%% CPU Cycles used: %ld\n", b);
	c = (double) b / (double)ITERATIONS;
	printf("%% Cycles per Call %f\n", c);
	wline();
	
	printf("%% Min\n");
	a = rdtsc();
	for(i = 0; i < ITERATIONS; ++i)
	{
			e = fmin(i, -i);
	}
	b = rdtsc() - a;
	printf("%% CPU Cycles used: %ld\n", b);
	c = (double) b / (double)ITERATIONS;
	printf("%% Cycles per Call %f\n", c);
	wline();
	
	printf("%% Ceil\n");
	a = rdtsc();
	for(i = 0; i < ITERATIONS; ++i)
	{
			e = ceil(i);
	}
	b = rdtsc() - a;
	printf("%% CPU Cycles used: %ld\n", b);
	c = (double) b / (double)ITERATIONS;
	printf("%% Cycles per Call %f\n", c);
	wline();
	
	printf("%% Floor\n");
	a = rdtsc();
	for(i = 0; i < ITERATIONS; ++i)
	{
			e = floor(i);
	}
	b = rdtsc() - a;
	printf("%% CPU Cycles used: %ld\n", b);
	c = (double) b / (double)ITERATIONS;
	printf("%% Cycles per Call %f\n", c);
	wline();
	
	printf("%% Abs\n");
	a = rdtsc();
	for(i = 0; i < ITERATIONS; ++i)
	{
			e = abs(i);
	}
	b = rdtsc() - a;
	printf("%% CPU Cycles used: %ld\n", b);
	c = (double) b / (double)ITERATIONS;
	printf("%% Cycles per Call %f\n", c);
	wline();
	
	printf("%% OwnSqrt\n");
	a = rdtsc();
	for(i = 0; i < ITERATIONS; ++i)
    {
        e = ownSqroot(i);
    }
	b = rdtsc() - a;
	printf("%% CPU Cycles used: %ld\n", b);
	c = (double) b / (double)ITERATIONS;
	printf("%% Cycles per Call %f\n", c);
	wline();
}

int smallif()
{
	int a = 12;

	if (a == 4)
	{
		return a;
	}
	else if (a == 8)
	{
		return a;
	}
	else if (a == 12)
	{
		return a;
	}	
	return a;
}

int smallswitch()
{
	int a = 12;
	switch(a)
	{
		case 4: return a;
		case 8: return a;
		case 12: return a;
		default: return a;
	}
}

int bigswitch()
{
	int a = 12;

	switch(a)
	{
		case 0: 
			return a;
			break;
		case 1: 
			return a;
			break;
		case 2: 
			return a;
			break;
		case 3: 
			return a;
			break;
		case 4: 
			return a;			
			break;
		case 5: 
			return a;			
			break;
		case 6: 
			return a;
			break;
		case 7: 
			return a;		
			break;
		case 8: 
			return a;
			break;
		case 9: 
			return a;
			break;
		case 10: 
			return a;
			break;
		case 11: 
			return a;
			break;
		case 12: 
			return a;
			break;
		default: 
			return a;
	}
}

int bigif()
{
	int a = 12;

	if (a == 0)
	{
		return a;
	}
	else if (a == 1)
	{		
		return a;
	}
	else if (a == 2)
	{		
		return a;
	}
	else if (a == 3)
	{		
		return a;
	}
	else if (a == 4)
	{		
		return a;
	}
	else if (a == 5)
	{		
		return a;
	}
	else if (a == 6)
	{		
		return a;
	}			
	else if (a == 7)
	{		
		return a;
	}
	else if (a == 8)
	{		
		return a;
	}
	else if (a == 9)
	{		
		return a;
	}
	else if (a == 10)
	{		
		return a;
	}
	else if (a == 11)
	{		
		return a;
	}
	else if (a == 12)
	{		
		return a;
	}

	return a;
}

int lotif()
{
	int a = 25;
	if (a == 0)
	{
		return a;
	}
	else if (a == 1)
	{		
		return a;
	}
	else if (a == 2)
	{		
		return a;
	}
	else if (a == 3)
	{		
		return a;
	}
	else if (a == 4)
	{		
		return a;
	}
	else if (a == 5)
	{		
		return a;
	}
	else if (a == 6)
	{		
		return a;
	}			
	else if (a == 7)
	{		
		return a;
	}
	else if (a == 8)
	{		
		return a;
	}
	else if (a == 9)
	{		
		return a;
	}
	else if (a == 10)
	{		
		return a;
	}
	else if (a == 11)
	{		
		return a;
	}
	else if (a == 12)
	{		
		return a;
	}
	else if (a == 13)
	{
		return a;
	}
	else if (a == 14)
	{		
		return a;
	}
	else if (a == 15)
	{		
		return a;
	}
	else if (a == 16)
	{		
		return a;
	}
	else if (a == 17)
	{		
		return a;
	}
	else if (a == 18)
	{		
		return a;
	}
	else if (a == 19)
	{		
		return a;
	}			
	else if (a == 20)
	{		
		return a;
	}
	else if (a == 21)
	{		
		return a;
	}
	else if (a == 22)
	{		
		return a;
	}
	else if (a == 23)
	{		
		return a;
	}
	else if (a == 24)
	{		
		return a;
	}
	else if (a == 25)
	{		
		return a;
	}
	return a;
}

int lotswitch()
{
	int a = 25;
	
	switch (a)
	{
		case 0:
			return a;
			break;
		case 1:
			return a;
			break;
		case 2:
			return a;
			break;
		case 3:
			return a;
			break;
		case 4:
			return a;
			break;
		case 5:
			return a;
			break;
		case 6:
			return a;
			break;
		case 7:
			return a;
			break;
		case 8:
			return a;
			break;
		case 9:
			return a;
			break;
		case 10:
			return a;
			break;
		case 11:
			return a;
			break;
		case 12:
			return a;
			break;
		case 13:
			return a;
			break;
		case 14:
			return a;
			break;
		case 15:
			return a;
			break;
		case 16:
			return a;
			break;
		case 17:
			return a;
			break;
		case 18:
			return a;
			break;
		case 19:
			return a;
			break;
		case 20:
			return a;
			break;
		case 21:
			return a;
			break;
		case 22:
			return a;
			break;
		case 23:
			return a;
			break;
		case 24:
			return a;
			break;
		case 25:
			return a;
			break;
		default: return a;
	}
}

void tMethods()
{
	long unsigned int i;
	int e = 12;
	int* f = (int *)12;
	
	int *g = (int *) malloc(10000000000000000 * sizeof(int));
	uint64_t a,b;
	double c;
	
	printf("%% Test Methods\n");
	
	wline();
		
	printf("%% Method\n");
	a = rdtsc();
    	
	for(i = 0; i < ITERATIONS; ++i)
    {
		method();
	}	
	
	b = rdtsc() - a;
	printf("%% CPU Cycles used: %ld\n", b);
	c = (double) b / (double)ITERATIONS;
	printf("%% Cycles per Call %f\n", c);
	wline();
	
	printf("%% Method with Return\n");
	a = rdtsc();
	
	for(i = 0; i < ITERATIONS; ++i)
    {
		methodRet();
	}
	
	b = rdtsc() - a;
	printf("%% CPU Cycles used: %ld\n", b);
	c = (double) b / (double)ITERATIONS;
	printf("%% Cycles per Call %f\n", c);
	wline();
	
	printf("%% Method with Parameters\n");
	a = rdtsc();
	
	for(i = 0; i < ITERATIONS; ++i)
    {
		methodParam(e);
	}
	
	b = rdtsc() - a;
	printf("%% CPU Cycles used: %ld\n", b);
	c = (double) b / (double)ITERATIONS;
	printf("%% Cycles per Call %f\n", c);
	wline();
	
	printf("%% Method mit Pointer as Param\n");
	a = rdtsc();
	
	for(i = 0; i < ITERATIONS; ++i)
    {
		methodPointer(f);
	}	
	
	b = rdtsc() - a;
	printf("%% CPU Cycles used: %ld\n", b);
	c = (double) b / (double)ITERATIONS;
	printf("%% Cycles per Call %f\n", c);
	wline();
	
	printf("%% Method with a Big [] as Param\n");
	a = rdtsc();
			
	for(i = 0; i < ITERATIONS; ++i)
    {
		methodBig(g);
	}
	
	b = rdtsc() - a;
	printf("%% CPU Cycles used: %ld\n", b);
	c = (double) b / (double)ITERATIONS;
	printf("%% Cycles per Call %f\n", c);
	wline();
	
	printf("%% Method with Param and Return\n");
	a = rdtsc();
		
	for(i = 0; i < ITERATIONS; ++i)
    {
		methodParamRet(e);
	}
	
	b = rdtsc() - a;
	printf("%% CPU Cycles used: %ld\n", b);
	c = (double) b / (double)ITERATIONS;
	printf("%% Cycles per Call %f\n", c);
	
	//free(g);
	
	line();
}

void tIf()
{
	uint64_t a, b;    
	double c;
	long i;
	
    printf("%% Test If\n");
	
	wline();

	printf("%% Small If\n");
	a = rdtsc();
    for (i = 0; i < ITERATIONS; ++i)
    {
		smallif();
    }
	b = rdtsc() - a;
	printf("%% CPU Cycles used: %ld\n", b);
	c = (double) b / (double)ITERATIONS;
	printf("%% Cycles per Call %f\n", c);
	wline();
	
	printf("%% Small Switch\n");
	a = rdtsc();
    for (i = 0; i < ITERATIONS; ++i)
    {
		smallswitch();
    }
	b = rdtsc() - a;
	printf("%% CPU Cycles used: %ld\n", b);
	c = (double) b / (double)ITERATIONS;
	printf("%% Cycles per Call %f\n", c);
	wline();
	
	printf("%% Big If\n");
	a = rdtsc();
    for(i = 0; i < ITERATIONS; ++i)
    {
        bigif();
    }
	b = rdtsc() - a;
	printf("%% CPU Cycles used: %ld\n", b);
	c = (double) b / (double)ITERATIONS;
	printf("%% Cycles per Call %f\n", c);
	wline();
	
	printf("%% Big Switch\n");
	a = rdtsc();
    for (i = 0; i < ITERATIONS; ++i)
    {
		bigswitch();
    }
	b = rdtsc() - a;
	printf("%% CPU Cycles used: %ld\n", b);
	c = (double) b / (double)ITERATIONS;
	printf("%% Cycles per Call %f\n", c);
	wline();
	
	printf("%% Lot If\n");
	a = rdtsc();
	for(i = 0; i < ITERATIONS; ++i)
    {
        lotif();
    }		
	b = rdtsc() - a;
	printf("%% CPU Cycles used: %ld\n", b);
	c = (double) b / (double)ITERATIONS;
	printf("%% Cycles per Call %f\n", c);
	wline();
	
	printf("%% Lot Switch\n");
	a = rdtsc();
    for (i = 0; i < ITERATIONS; ++i)
    {
		lotswitch();
    }
	b = rdtsc() - a;
	printf("%% CPU Cycles used: %ld\n", b);
	c = (double) b / (double)ITERATIONS;
	printf("%% Cycles per Call %f\n", c);
	wline();
}

// Method for calculations
void tCalculations()
{
	uint64_t a,b,i;
	float x,y,z  = 1.0;
	y = 12.0;
	int q,r,s = 1;
	double c;
	
	q = 1;
	r = q;
	
	printf("%% Calculations\n");
	
	wline();
	printf("%% + int\n");
	a = rdtsc();
    
	for(i = 0; i < ITERATIONS; ++i)
    {
       r = r + s;
    }
		
	b = rdtsc() - a;
	printf("%% CPU Cycles used: %ld\n", b);
	c = (double) b / (double)ITERATIONS;
	printf("%% Cycles per Call %f\n", c);
	
	wline();	
	printf("%% + double\n");
	a = rdtsc();
    
	for(i = 0; i < ITERATIONS; ++i)
    {
       x = x + y;
    }
		
	b = rdtsc() - a;
	printf("%% CPU Cycles used: %ld\n", b);
	c = (double) b / (double)ITERATIONS;
	printf("%% Cycles per Call %f\n", c);
	
	wline();
	printf("%% - int\n");
	a = rdtsc();
    
	for(i = 0; i < ITERATIONS; ++i)
    {
        r = r - s;
    }
		
	b = rdtsc() - a;
	printf("%% CPU Cycles used: %ld\n", b);
	c = (double) b / (double)ITERATIONS;
	printf("%% Cycles per Call %f\n", c);
	
	wline();	
	printf("%% - double\n");
	a = rdtsc();
    
	for(i = 0; i < ITERATIONS; ++i)
    {
       x = x - y;
    }
		
	b = rdtsc() - a;
	printf("%% CPU Cycles used: %ld\n", b);
	c = (double) b / (double)ITERATIONS;
	printf("%% Cycles per Call %f\n", c);
	
	wline();
	printf("%% * int\n");
	a = rdtsc();
    
	for(i = 0; i < ITERATIONS; ++i)
    {
        r = r * s;
    }
		
	b = rdtsc() - a;
	printf("%% CPU Cycles used: %ld\n", b);
	c = (double) b / (double)ITERATIONS;
	printf("%% Cycles per Call %f\n", c);
	
	wline();
	printf("%% * double\n");
	a = rdtsc();
    
	for(i = 0; i < ITERATIONS; ++i)
    {
        x = x * y;
    }
		
	b = rdtsc() - a;
	printf("%% CPU Cycles used: %ld\n", b);
	c = (double) b / (double)ITERATIONS;
	printf("%% Cycles per Call %f\n", c);
	
	wline();
	printf("%% / int\n");
	a = rdtsc();
    
	for(i = 0; i < ITERATIONS; ++i)
    {
      q = r / s;
    }
		
	b = rdtsc() - a;
	printf("%% CPU Cycles used: %ld\n", b);
	c = (double) b / (double)ITERATIONS;
	printf("%% Cycles per Call %f\n", c);
	
	wline();
	printf("%% / double\n");
	a = rdtsc();
    
	for(i = 0; i < ITERATIONS; ++i)
    {
       x = x / y;
    }
		
	b = rdtsc() - a;
	printf("%% CPU Cycles used: %ld\n", b);
	c = (double) b / (double)ITERATIONS;
	printf("%% Cycles per Call %f\n", c);
	
	wline();
	printf("%% x * y * z * x\n");
	a = rdtsc();
    
	for(i = 0; i < ITERATIONS; ++i)
    {
       x = x * y * z * x;
    }
		
	b = rdtsc() - a;
	printf("%% CPU Cycles used: %ld\n", b);
	c = (double) b / (double)ITERATIONS;
	printf("%% Cycles per Call %f\n", c);
	
	wline();
	printf("%% (x * y) * (z * x)\n");
	a = rdtsc();
    
	for(i = 0; i < ITERATIONS; ++i)
    {
       x = (x * y) * (z * x);
    }
		
	b = rdtsc() - a;
	printf("%% CPU Cycles used: %ld\n", b);
	c = (double) b / (double)ITERATIONS;
	printf("%% Cycles per Call %f\n", c);
	
	wline();
	printf("%% / double\n");
	a = rdtsc();
    
	for(i = 0; i < ITERATIONS; ++i)
    {
       x = x / y;
    }
		
	b = rdtsc() - a;
	printf("%% CPU Cycles used: %ld\n", b);
	c = (double) b / (double)ITERATIONS;
	printf("%% Cycles per Call %f\n", c);

	line();
}

int main()
{
  line();
  printf("%% Runtime Cost Benchmark                   %%\n");
  printf("%% Seminar: Effiziente Programmierung in C  %%\n");
  line();
  printf("%% Iterations per Call: %d            %%\n", ITERATIONS);
  printf("%% Clocks per Sec: %ld                  %%\n", CLOCKS_PER_SEC);
  line();

  tMethods();        
		
  tCalculations();
		
  mathFunctions();
  
  tIf();	
		
  line();
  line();
	
  return 0;
}
